
//		Snapshot.cpp - console buffer save/load plugin for TCC


#define UNICODE							1
#define _UNICODE						1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>
#include <mmsystem.h>

#include "PlugIn.h"
#include "TakeCmd.h"


#define VER_MAJOR						1
#define VER_MINOR						2
#define VER_BUILD						0
#define VER_PATCH						3

#ifdef _WIN64
#define PLATFORM						L"x64"
#else	//	_WIN64
#define PLATFORM						L"x86"
#endif	//	_WIN64

#define PLUGIN_NAME						L"Snapshot"


#define SHORT_BUF_LEN					1024

#define MAX_ARG_LEN						8192
#define MAX_ARGS						1024
#define MAX_FILENAME_LEN				16384

#define FILE_BUFFER_SIZE				8192

#define ERRHANDLE						GetStdHandle( STD_ERROR_HANDLE )

#define SPACE							0x20


#define SNAP_MODE_SAVE					0
#define SNAP_MODE_RESTORE				1
#define SNAP_MODE_INFO					2

#define SNAP_FLAG_OVERWRITE				0x0001
#define SNAP_FLAG_NO_TIME_STAMPS		0x0002
#define SNAP_FLAG_SKIP_CHECKSUM			0x0004
#define SNAP_FLAG_IGNORE_OTHERS			0x0008
#define SNAP_FLAG_NO_FOOTERS			0x0010
#define SNAP_FLAG_NO_ZERO_FOOTERS		0x0020
#define SNAP_FLAG_QUIET					0x0080

#define RECURSE_FLAG_RECURSE			0x8000
#define RECURSE_FLAG_NO_FNF_ERROR		0x0001
#define RECURSE_FLAG_NO_HIDDEN_DIRS		0x0002
#define RECURSE_FLAG_NO_SYSTEM_DIRS		0x0004
#define RECURSE_FLAG_NO_JUNCTIONS		0x0008



/***************************************************************

           Snapshot file format:

 Word      Data
  0-3      "Snapshot"  (8 bytes ASCII)
    4      File version, always 1
    5      Year
    6      Month (high) and date (low)
    7      Hour (high) and minute (low)
    8      Day of the week (high); seconds (low)
 9-10      UTC offset
   11      Console columns
   12      Console rows
   13      Cursor column
   14      Cursor row
   15      Default color
16-17      Console code page (input)
18-19      Console code page (output)
   20      Window height (rows)
   21      Unused; zero
22-23      CRC32 of raw data
24-39      Computer name (16 characters Unicode)

  40-      Data words (screen codes), OR:

		   0xFFF9 count data (3 words):  repeat value "data" count times;
			   if count = 0, repeat 65536 times

		   0xFFFA count (2 words): repeat value 0x0020 count times;
			   if count = 0, repeat 65536 times

		   0xFFFB data (2 words):  escape following character; treat literally

		   0xFFFC data:  use screen attribute "data" for following characters

		   0xFFFD:  marks end of data; if not at end of buffer, fill to end
		       with 0x0020 in current color


***************************************************************/



struct _SnapHeader {
	char Signature[8];
	WORD Version;
	WORD Year;
	WORD Date;
	WORD Time;
	WORD Seconds;
	WORD BiasLow;
	WORD BiasHigh;
	WORD NumColumns;
	WORD NumRows;
	WORD CursorColumn;
	WORD CursorRow;
	WORD Color;
	DWORD CodePageInput;
	DWORD CodePageOutput;
	WORD WindowHeight;
	WORD Unused;
	DWORD CRC32;
	wchar_t ComputerName[16];
};



#define RepeatCode				0xfff9
#define RepeatSpaceCode			0xfffa
#define EscapeCode				0xfffb
#define ColorCode				0xfffc
#define EndOfDataCode			0xfffd


#define VEFConCols				0
#define VEFConRows				1
#define VEFCurCol				2
#define VEFCurRow				3
#define VEFColor				4
#define VEFWinLeft				5
#define VEFWinTop				6
#define VEFWinRight				7
#define VEFWinBottom			8
#define VEFWinPosLeft			9
#define VEFWinPosTop			10
#define VEFHeaderEnd			11



HINSTANCE MyDLLHandle			= NULL;

LPTSTR LastFilename				= NULL;


const wchar_t ErrorString[] = L"*ERROR*";
const char Signature[]		= "Snapshot";

const wchar_t PluginRegKey[]	= L"SOFTWARE\\JPPlugins\\Snapshot";
const wchar_t QuietRegValue[]	= L"Quiet";




/* CRC-32 Lookup Table */ 
DWORD crc32_table[256] =
{
  /*   0 -- */  0x00000000, 0x77073096, 0xEE0E612C, 0x990951BA, 
  /*   4 -- */  0x076DC419, 0x706AF48F, 0xE963A535, 0x9E6495A3, 
  /*   8 -- */  0x0EDB8832, 0x79DCB8A4, 0xE0D5E91E, 0x97D2D988, 
  /*  12 -- */  0x09B64C2B, 0x7EB17CBD, 0xE7B82D07, 0x90BF1D91, 
  /*  16 -- */  0x1DB71064, 0x6AB020F2, 0xF3B97148, 0x84BE41DE, 
  /*  20 -- */  0x1ADAD47D, 0x6DDDE4EB, 0xF4D4B551, 0x83D385C7, 
  /*  24 -- */  0x136C9856, 0x646BA8C0, 0xFD62F97A, 0x8A65C9EC, 
  /*  28 -- */  0x14015C4F, 0x63066CD9, 0xFA0F3D63, 0x8D080DF5, 
  /*  32 -- */  0x3B6E20C8, 0x4C69105E, 0xD56041E4, 0xA2677172, 
  /*  36 -- */  0x3C03E4D1, 0x4B04D447, 0xD20D85FD, 0xA50AB56B, 
  /*  40 -- */  0x35B5A8FA, 0x42B2986C, 0xDBBBC9D6, 0xACBCF940, 
  /*  44 -- */  0x32D86CE3, 0x45DF5C75, 0xDCD60DCF, 0xABD13D59, 
  /*  48 -- */  0x26D930AC, 0x51DE003A, 0xC8D75180, 0xBFD06116, 
  /*  52 -- */  0x21B4F4B5, 0x56B3C423, 0xCFBA9599, 0xB8BDA50F, 
  /*  56 -- */  0x2802B89E, 0x5F058808, 0xC60CD9B2, 0xB10BE924, 
  /*  60 -- */  0x2F6F7C87, 0x58684C11, 0xC1611DAB, 0xB6662D3D, 
  /*  64 -- */  0x76DC4190, 0x01DB7106, 0x98D220BC, 0xEFD5102A, 
  /*  68 -- */  0x71B18589, 0x06B6B51F, 0x9FBFE4A5, 0xE8B8D433, 
  /*  72 -- */  0x7807C9A2, 0x0F00F934, 0x9609A88E, 0xE10E9818, 
  /*  76 -- */  0x7F6A0DBB, 0x086D3D2D, 0x91646C97, 0xE6635C01, 
  /*  80 -- */  0x6B6B51F4, 0x1C6C6162, 0x856530D8, 0xF262004E, 
  /*  84 -- */  0x6C0695ED, 0x1B01A57B, 0x8208F4C1, 0xF50FC457, 
  /*  88 -- */  0x65B0D9C6, 0x12B7E950, 0x8BBEB8EA, 0xFCB9887C, 
  /*  92 -- */  0x62DD1DDF, 0x15DA2D49, 0x8CD37CF3, 0xFBD44C65, 
  /*  96 -- */  0x4DB26158, 0x3AB551CE, 0xA3BC0074, 0xD4BB30E2, 
  /* 100 -- */  0x4ADFA541, 0x3DD895D7, 0xA4D1C46D, 0xD3D6F4FB, 
  /* 104 -- */  0x4369E96A, 0x346ED9FC, 0xAD678846, 0xDA60B8D0, 
  /* 108 -- */  0x44042D73, 0x33031DE5, 0xAA0A4C5F, 0xDD0D7CC9, 
  /* 112 -- */  0x5005713C, 0x270241AA, 0xBE0B1010, 0xC90C2086, 
  /* 116 -- */  0x5768B525, 0x206F85B3, 0xB966D409, 0xCE61E49F, 
  /* 120 -- */  0x5EDEF90E, 0x29D9C998, 0xB0D09822, 0xC7D7A8B4, 
  /* 124 -- */  0x59B33D17, 0x2EB40D81, 0xB7BD5C3B, 0xC0BA6CAD, 
  /* 128 -- */  0xEDB88320, 0x9ABFB3B6, 0x03B6E20C, 0x74B1D29A, 
  /* 132 -- */  0xEAD54739, 0x9DD277AF, 0x04DB2615, 0x73DC1683, 
  /* 136 -- */  0xE3630B12, 0x94643B84, 0x0D6D6A3E, 0x7A6A5AA8, 
  /* 140 -- */  0xE40ECF0B, 0x9309FF9D, 0x0A00AE27, 0x7D079EB1, 
  /* 144 -- */  0xF00F9344, 0x8708A3D2, 0x1E01F268, 0x6906C2FE, 
  /* 148 -- */  0xF762575D, 0x806567CB, 0x196C3671, 0x6E6B06E7, 
  /* 152 -- */  0xFED41B76, 0x89D32BE0, 0x10DA7A5A, 0x67DD4ACC, 
  /* 156 -- */  0xF9B9DF6F, 0x8EBEEFF9, 0x17B7BE43, 0x60B08ED5, 
  /* 160 -- */  0xD6D6A3E8, 0xA1D1937E, 0x38D8C2C4, 0x4FDFF252, 
  /* 164 -- */  0xD1BB67F1, 0xA6BC5767, 0x3FB506DD, 0x48B2364B, 
  /* 168 -- */  0xD80D2BDA, 0xAF0A1B4C, 0x36034AF6, 0x41047A60, 
  /* 172 -- */  0xDF60EFC3, 0xA867DF55, 0x316E8EEF, 0x4669BE79, 
  /* 176 -- */  0xCB61B38C, 0xBC66831A, 0x256FD2A0, 0x5268E236, 
  /* 180 -- */  0xCC0C7795, 0xBB0B4703, 0x220216B9, 0x5505262F, 
  /* 184 -- */  0xC5BA3BBE, 0xB2BD0B28, 0x2BB45A92, 0x5CB36A04, 
  /* 188 -- */  0xC2D7FFA7, 0xB5D0CF31, 0x2CD99E8B, 0x5BDEAE1D, 
  /* 192 -- */  0x9B64C2B0, 0xEC63F226, 0x756AA39C, 0x026D930A, 
  /* 196 -- */  0x9C0906A9, 0xEB0E363F, 0x72076785, 0x05005713, 
  /* 200 -- */  0x95BF4A82, 0xE2B87A14, 0x7BB12BAE, 0x0CB61B38, 
  /* 204 -- */  0x92D28E9B, 0xE5D5BE0D, 0x7CDCEFB7, 0x0BDBDF21, 
  /* 208 -- */  0x86D3D2D4, 0xF1D4E242, 0x68DDB3F8, 0x1FDA836E, 
  /* 212 -- */  0x81BE16CD, 0xF6B9265B, 0x6FB077E1, 0x18B74777, 
  /* 216 -- */  0x88085AE6, 0xFF0F6A70, 0x66063BCA, 0x11010B5C, 
  /* 220 -- */  0x8F659EFF, 0xF862AE69, 0x616BFFD3, 0x166CCF45, 
  /* 224 -- */  0xA00AE278, 0xD70DD2EE, 0x4E048354, 0x3903B3C2, 
  /* 228 -- */  0xA7672661, 0xD06016F7, 0x4969474D, 0x3E6E77DB, 
  /* 232 -- */  0xAED16A4A, 0xD9D65ADC, 0x40DF0B66, 0x37D83BF0, 
  /* 236 -- */  0xA9BCAE53, 0xDEBB9EC5, 0x47B2CF7F, 0x30B5FFE9, 
  /* 240 -- */  0xBDBDF21C, 0xCABAC28A, 0x53B39330, 0x24B4A3A6, 
  /* 244 -- */  0xBAD03605, 0xCDD70693, 0x54DE5729, 0x23D967BF, 
  /* 248 -- */  0xB3667A2E, 0xC4614AB8, 0x5D681B02, 0x2A6F2B94, 
  /* 252 -- */  0xB40BBE37, 0xC30C8EA1, 0x5A05DF1B, 0x2D02EF8D
};



const wchar_t SnapHelpText[]	= L"Save, restore, or show information about console snapshot files.  \n\
\n\
SNAP /I /N: /O /Q /R /S filename  \r\n\
\n\
\t/I\tdisplay information about snapshot files\n\
\t/N:\tdisable features\n\
\t/O\toverwrite an existing file\n\
\t/Q\tquietly\n\
\t/R\treload the console buffer from a snapshot file  \n\
\t/S\tsearch into subdirectories\n\
\n\
If neither /I nor /R is given, SNAP saves the console buffer  \n\
to a file.  The filename is optional.\n\
\n\
If /I is given, you can use wildcards, multiple filenames, or /S\n\
to recurse.\n";


const wchar_t _LastSnapHelpText[] = L"%_LASTSNAP\n\
\n\
Returns the filename of the last snapshot.  If any error was\n\
detected, returns the string *ERROR*.  If the SNAP command  \n\
has not yet been called, returns an empty string.\n";


const wchar_t PluginFeatures[] = L"snap,_lastsnap";



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( ERRHANDLE, shortbuf );

	return 0;
}



int GetErrorText( int err, LPTSTR outString )
{
	if ( !outString )
		return -1;

	if ( err == -1 )
		wcscpy_s( outString, SHORT_BUF_LEN, L"Unknown error" );
	else if ( FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK, NULL, err, 0, outString, SHORT_BUF_LEN, NULL ) == 0 )
		Sprintf( outString, L"Error %d", err );

	size_t l = wcslen( outString );

	if ( l > 0 )
		while ( iswspace( outString[l - 1] ) ) {
			l--;
			outString[l] = '\0';
		}

	return 0;
}



int GetLastErrorText( LPTSTR outString )
{
	return GetErrorText( GetLastError(), outString );
}



#include "ParseArgs.cpp"



LPTSTR _PathPart( LPTSTR Filename, LPTSTR outPath )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;


	BOOL FoundSlash			= FALSE;
	size_t EndOfPathPart	= -1;


	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) ) {
			EndOfPathPart = i;
			FoundSlash = TRUE;
		}
		else if ( ( Filename[i] == ':' ) && !FoundSlash )
			EndOfPathPart = i;
	}

	if ( EndOfPathPart == -1 )
		return NULL;

	if ( outPath ) {
		for ( size_t i = 0; i <= EndOfPathPart; i++ )
			outPath[i] = Filename[i];

		outPath[EndOfPathPart + 1] = '\0';
		return outPath;
	}

	return Filename;
}



LPTSTR _NamePart( LPTSTR Filename, LPTSTR outFilename )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;


	BOOL FoundSlash			= FALSE;
	size_t EndOfPathPart	= -1;


	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) ) {
			EndOfPathPart = i;
			FoundSlash = TRUE;
		}
		else if ( ( Filename[i] == ':' ) && !FoundSlash )
			EndOfPathPart = i;
	}

	if ( EndOfPathPart == -1 ) {
		if ( outFilename ) {
			wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename );
			return outFilename;
		}
		else
			return Filename;
	}

	if ( outFilename ) {
		wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename + EndOfPathPart + 1 );
		return outFilename;
	}
	else
		return Filename + EndOfPathPart + 1;
}



int AppendFilename( LPTSTR Path, LPTSTR Filename, size_t BufferSize )
{
	if ( !Path || !Filename )
		return -1;

	
	BOOL NeedSlash		= FALSE;

	for ( size_t i = 0; Path[i]; i++ ) {
		if ( ( Path[i] == '/' ) || ( Path[i] == '\\' ) || ( Path[i] == ':' ) )
			NeedSlash = FALSE;
		else
			NeedSlash = TRUE;
	}

	if ( ( Filename[0] == '/' ) || ( Filename[0] == '\\' ) )
		NeedSlash = FALSE;
	
	if ( NeedSlash )
		wcscat_s( Path, BufferSize, L"\\" );

	wcscat_s( Path, BufferSize, Filename );
	return 0;
}



BOOL IsColonOrEquals( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '=' ) );
}



DWORD ComputeCRC32( WORD *data, long int size )
{
	DWORD crc = 0xffffffff;
	int k = 0;

	for ( long int i = 0; i < size; i++ ) {
		k = ( crc ^ ( data[i] & 0x00ff ) ) & 0x000000ff;
		crc = ( ( crc >> 8 ) & 0x00ffffff ) ^ crc32_table[k];

		k = ( crc ^ ( data[i] >> 8 ) ) & 0x000000ff;
		crc = ( ( crc >> 8 ) & 0x00ffffff ) ^ crc32_table[k];
	}

	return ~crc;
}



DWORD RegGetDWord( HKEY KeyHandle, LPCTSTR ValueName, DWORD Min, DWORD Max, DWORD Default, int *Error )
{
	DWORD value			= -1;
	DWORD bufsize		= sizeof( DWORD );
	DWORD valtype		= 0;


	RegQueryValueEx( KeyHandle, ValueName, NULL, &valtype, (BYTE*) &value, &bufsize );

	if ( ( valtype != REG_DWORD ) || ( value < Min ) || ( value > Max ) ) {
		value = Default;

		if ( Error )
			*Error += 1;
	}

	return value;
}



BOOL SilencedByRegistry()
{
	HKEY Key			= NULL;
	DWORD value			= 0;
	int err				= 0;

	BOOL rv				= FALSE;


	if ( RegOpenKey( HKEY_CURRENT_USER, PluginRegKey, &Key ) == ERROR_SUCCESS ) {
		value = RegGetDWord( Key, QuietRegValue, 0, UINT_MAX, 0, &err );

		if ( value && !err )
			rv = TRUE;

		RegCloseKey( Key );
		Key = NULL;
	}

	if ( err ) {
		err = 0;

		if ( RegOpenKey( HKEY_LOCAL_MACHINE, PluginRegKey, &Key ) == ERROR_SUCCESS ) {
			value = RegGetDWord( Key, QuietRegValue, 0, UINT_MAX, 0, &err );

			if ( value && !err )
				rv = TRUE;

			RegCloseKey( Key );
			Key = NULL;
		}
	}

	return rv;
}



DWORD ChecksumTwoBuffers( WORD *Chars, WORD *Attribs, unsigned int Size )
{
	DWORD crc		= 0xffffffff;
	int k			= 0;


	for ( unsigned int i = 0; i < Size; i++ ) {
		k = ( crc ^ ( Chars[i] & 0x00ff ) ) & 0x000000ff;
		crc = ( ( crc >> 8 ) & 0x00ffffff ) ^ crc32_table[k];

		k = ( crc ^ ( Chars[i] >> 8 ) ) & 0x000000ff;
		crc = ( ( crc >> 8 ) & 0x00ffffff ) ^ crc32_table[k];

		k = ( crc ^ ( Attribs[i] & 0x00ff ) ) & 0x000000ff;
		crc = ( ( crc >> 8 ) & 0x00ffffff ) ^ crc32_table[k];

		k = ( crc ^ ( Attribs[i] >> 8 ) ) & 0x000000ff;
		crc = ( ( crc >> 8 ) & 0x00ffffff ) ^ crc32_table[k];
	}

	return ~crc;
}



long int RepeatCheck( WORD *buffer, long int size, long int index )
{
	long int j = 2;
	WORD data  = buffer[index];
	WORD color = buffer[index + 1];
	
	do {
		if ( ( ( index + j ) > ( size - 2 ) ) && ( data == 0x20 ) )
			return -1;

		if ( ( j > ( 2 * 65536 ) ) && ( data != 0x20 ) )
			break;

		if ( ( index + j ) > ( size - 2 ) )
			break;

		if ( ( buffer[index + j] != data ) || ( buffer[index + j + 1] != color ) )
			break;

		j += 2;

	} while ( TRUE );
	

	j /= 2;
	
	if ( j > 65536 )
		j = 65536;

	return j;
}



int RepeatCheck( WORD *CharBuffer, WORD *AttrBuffer, unsigned int BufferSize, unsigned int Index )
{
	int j			= 1;
	WORD data		= CharBuffer[Index];
	WORD color		= AttrBuffer[Index];


	do {
		if ( ( ( Index + j ) > ( BufferSize - 1 ) ) && ( data == SPACE ) )
			return -1;

		if ( ( j > 65536 ) && ( data != SPACE ) )
			break;

		if ( ( Index + j ) > ( BufferSize - 1 ) )
			break;

		if ( ( CharBuffer[Index + j] != data ) || ( AttrBuffer[Index + j] != color ) )
			break;

		j++;
	}
	while ( TRUE );

	if ( j > 65536 )
		j = 65536;

	return j;
}



int FlushBuffer( WORD *Buffer, unsigned int *BufferPtr, unsigned int BufferSize, HANDLE FileHandle )
{
	DWORD BytesWritten		= 0;
	int rv					= 0;


	if ( *BufferPtr == 0 )
		return 0;


	rv = WriteFile( FileHandle, Buffer, *BufferPtr * sizeof( WORD ), &BytesWritten, NULL );
	*BufferPtr = 0;

	if ( rv )
		return 0;
	
	return 2;
}



int WriteWord( WORD Data, WORD *Buffer, unsigned int *BufferPtr, unsigned int BufferSize, HANDLE FileHandle )
{
	int rv				= 0;


	if ( *BufferPtr >= BufferSize )
		if ( rv = FlushBuffer( Buffer, BufferPtr, BufferSize, FileHandle ) )
			return rv;

	Buffer[*BufferPtr] = Data;
	*BufferPtr += 1;
	return 0;
}



int ReadBuffer( WORD *Buffer, unsigned int *BufferPtr, unsigned int *BufferValid, unsigned int BufferSize, HANDLE FileHandle )
{
	DWORD BytesRead		= 0;
	int rv				= 0;


	*BufferPtr		= 0;
	*BufferValid	= 0;

	if ( !ReadFile( FileHandle, Buffer, BufferSize * sizeof( WORD ), &BytesRead, NULL ) )
		return -1;

	if ( !BytesRead )
		return -2;

	*BufferValid = BytesRead / 2;
	return 0;
}



int ReadWord( WORD *Buffer, unsigned int *BufferPtr, unsigned int *BufferValid, unsigned int BufferSize, HANDLE FileHandle )
{
	int rv			= 0;


	if ( *BufferPtr >= *BufferValid ) {
		rv = ReadBuffer( Buffer, BufferPtr, BufferValid, BufferSize, FileHandle );

		if ( rv )
			return rv;
	}

	rv = Buffer[*BufferPtr];
	*BufferPtr += 1;
	return rv;
}



int ResizeConsoleWindow( HANDLE ConsoleHandle, HWND WindowHandle, int NewColumns, int NewRows )
{
	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
	RECT WindowSize;
	CONSOLE_FONT_INFO FontInfo;

	int oldcols				= 0;
	int oldrows				= 0;
	long int oldwidth		= 0;
	long int oldheight		= 0;
	long int newwidth		= 0;
	long int newheight		= 0;
	long int OverheadX		= 0;
	long int OverheadY		= 0;

	int hscrollbar			= GetSystemMetrics( SM_CXHSCROLL );
	int vscrollbar			= GetSystemMetrics( SM_CXVSCROLL );

	int rv					= 0;


	//		Get the current window size in rows and columns :

	GetConsoleScreenBufferInfo( ConsoleHandle, &ConsoleInfo );
	oldcols = ConsoleInfo.srWindow.Right - ConsoleInfo.srWindow.Left + 1;
	oldrows = ConsoleInfo.srWindow.Bottom - ConsoleInfo.srWindow.Top + 1;

	if ( ( oldcols == NewColumns ) && ( oldrows == NewRows ) )
		return 0;

	//		Get the current window size in pixels :

	GetWindowRect( WindowHandle, &WindowSize );
	oldwidth  = WindowSize.right - WindowSize.left;
	oldheight = WindowSize.bottom - WindowSize.top;

	//		Get the console's font size :

	GetCurrentConsoleFont( ConsoleHandle, FALSE, &FontInfo );
	FontInfo.dwFontSize = GetConsoleFontSize( ConsoleHandle, FontInfo.nFont );

	//		Figure the overhead used by the non-client area of the window,
	//		not including any scrollbars :

	OverheadX = oldwidth - ( oldcols * FontInfo.dwFontSize.X );
	if ( ConsoleInfo.dwSize.Y > oldrows )
		OverheadX -= GetSystemMetrics( SM_CXVSCROLL );

	OverheadY = oldheight - ( oldrows * FontInfo.dwFontSize.Y );

	//		Calculate the new window size in pixels :

	newwidth  = ( NewColumns * FontInfo.dwFontSize.X ) + OverheadX + ( ( ConsoleInfo.dwSize.Y > NewRows ) ? vscrollbar : 0 );
	newheight = ( NewRows * FontInfo.dwFontSize.Y ) + OverheadY + ( ( ConsoleInfo.dwSize.X > NewColumns ) ? hscrollbar : 0 );

	//		Resize the console window :

	MoveWindow( WindowHandle, WindowSize.left, WindowSize.top, newwidth, newheight, TRUE );
	Sleep( 10 );
	return 0;
}



BOOL inrange( WORD value, WORD range1, WORD range2 )
{
	WORD a = ( range1 <= range2 ) ? range1 : range2;
	WORD b = ( range1 <= range2 ) ? range2 : range1;

	return ( ( value >= a ) && ( value <= b ) );
}



int CheckHeader( _SnapHeader *Header )
{
	//		Returns 1 for a Snapshot file, 2 for a Vincent Fatica
	//		SaveCon file, or 0 for anything else.


	WORD *HeaderWords		= (WORD*) Header;


	if ( !memcmp( Header->Signature, Signature, 8 ) ) {

		//		Signature matches; do a few additional sanity checks :

		if ( !inrange( Header->NumColumns, 22, 512 ) )
			return 0;

		if ( !inrange( Header->NumRows, 8, 9999 ) )
			return 0;

		if ( ( Header->CursorColumn >= Header->NumColumns ) || ( Header->CursorRow >= Header->NumRows ) )
			return 0;

		return 1;
	}

	//		No signature -- it's not a Snapshot file.
	//		It *might* be a VEF SaveCon file, so do a whole bunch of sanity checks :

	if ( !inrange( HeaderWords[VEFConCols], 22, 512 ) )
		return 0;

	if ( !inrange( HeaderWords[VEFConRows], 8, 9999 ) )
		return 0;

	if ( ( HeaderWords[VEFCurCol] >= HeaderWords[VEFConCols] ) || ( HeaderWords[VEFCurRow] >= HeaderWords[VEFConRows] ) )
		return 0;

	if ( ( ( HeaderWords[VEFColor] & 0xf0 ) >> 4 ) == ( HeaderWords[VEFColor] & 0x0f ) )
		return 0;

	if ( ( HeaderWords[VEFWinRight] <= HeaderWords[VEFWinLeft] ) || ( HeaderWords[VEFWinBottom] <= HeaderWords[VEFWinTop] ) )
		return 0;

	if ( ( ( HeaderWords[VEFWinRight] - HeaderWords[VEFWinLeft] ) < 21 ) || ( ( HeaderWords[VEFWinBottom] - HeaderWords[VEFWinTop] ) < 4 ) )
		return 0;

	if ( ( ( HeaderWords[VEFWinRight] - HeaderWords[VEFWinLeft] ) >= HeaderWords[VEFConCols] ) || ( ( HeaderWords[VEFWinBottom] - HeaderWords[VEFWinTop] ) >= HeaderWords[VEFConRows] ) )
		return 0;

	return 2;
}



/*
int GetYorN( int default )
{
	int inKey = 0;

	do {
		inKey = GetKeystroke( 0x210 );
	} while ( ( inKey != 'Y' ) && ( inKey != 'N' ) && ( inKey != '\r' ) && ( inKey != '\x1b' ) );

	if ( inKey == '\x1b' )
		inKey = 'N';
	else if ( inKey == '\r' ) {
		if ( default )
			inKey = 'Y';
		else
			inKey = 'N';
	}

	Printf( L"%c\r\n", inKey );

	return ( inKey == 'Y' ) ? 1 : 0;
}
*/



int DisplaySnapInfo( LPTSTR Filename, BOOL IgnoreNonSnapshotFiles )
{
	if ( !Filename )
		return -1;

	if ( Filename[0] == '\0' )
		return -1;


	wchar_t ErrorText[SHORT_BUF_LEN]		= L"";

	HANDLE FileHandle			= NULL;
	DWORD BytesRead				= 0;

	_SnapHeader Header;
	WORD *HeaderWords			= (WORD*) &Header;

	int FileType				= 0;


	int rv = 0;


	//		Open file handle :

	FileHandle = CreateFile( Filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

	if ( !FileHandle || ( FileHandle == INVALID_HANDLE_VALUE ) ) {
		GetLastErrorText( ErrorText );
		QPuts( L"\r\n" );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error opening \"%s\":\r\n", Filename );
		Qprintf( ERRHANDLE, L"    %s\r\n", ErrorText );
		return -2;
	}

	//		Read header from file :

	memset( &Header, 0x00, sizeof( _SnapHeader ) );

	rv = ReadFile( FileHandle, &Header, sizeof( _SnapHeader ), &BytesRead, NULL );
	CloseHandle( FileHandle );

	if ( !rv ) {
		rv = GetLastError();
		if ( !rv )
			rv = -1;

		GetErrorText( rv, ErrorText );
		QPuts( L"\r\n" );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error reading \"%s\":\r\n", Filename );
		Qprintf( ERRHANDLE, L"    %s\r\n", ErrorText );
		return -2;
	}
	else if ( BytesRead != sizeof( _SnapHeader ) ) {
		if ( !IgnoreNonSnapshotFiles )
			Printf( L"\r\n\"%s\" is not a Snapshot file.\r\n", Filename );

		return 0;
	}


	//		Check the file header :

	FileType = CheckHeader( &Header );

	switch ( FileType ) {

		case 1 :

			Printf( L"\r\n\"%s\" - Snapshot v%d save file\r\n", Filename, Header.Version );
			QPuts( L"   Taken              : ");

			if ( Header.Year ) {
				int HourOffset		= - ( ( Header.BiasHigh << 16 ) | Header.BiasLow );
				int MinuteOffset = HourOffset % 60;
				if ( MinuteOffset < 0 )
					MinuteOffset = -MinuteOffset;
				HourOffset = HourOffset / 60;
				Printf( L"%04d-%02d-%02d %02d:%02d:%02d UTC%s%d:%02d\r\n", Header.Year, Header.Date >> 8, Header.Date & 0xff, Header.Time >> 8, Header.Time & 0xff, Header.Seconds & 0xff, ( HourOffset >= 0 ) ? L"+" : L"", HourOffset, MinuteOffset );
			}
			else
				QPuts( L"not available\r\n" );

			Printf( L"   On computer        : %s\r\n", Header.ComputerName[0] ? Header.ComputerName : L"not available" );
			Printf( L"   Checksum           : %08X\r\n", Header.CRC32 );
			Printf( L"   Screen buffer size : %d rows x %d columns\r\n", Header.NumRows, Header.NumColumns );
			Printf( L"   Cursor position    : row %d, column %d\r\n", Header.CursorRow, Header.CursorColumn );
			Printf( L"   Screen attribute   : %04x\r\n", Header.Color & 0x00ff );
			Printf( L"   Window height      : %d rows\r\n", Header.WindowHeight );

			if ( Header.CodePageInput == Header.CodePageOutput )
				Printf( L"   Code page          : %d\r\n", Header.CodePageInput );
			else
				Printf( L"   Code page          : %d (input) / %d (output)\r\n", Header.CodePageInput, Header.CodePageOutput );

			break;

		case 2 :

			Printf( L"\r\n\"%s\" - Vincent Fatica SAVECON file\r\n", Filename );
			Printf( L"   Screen buffer size : %d rows x %d columns\r\n", HeaderWords[VEFConRows], HeaderWords[VEFConCols] );
			Printf( L"   Cursor position    : row %d, column %d\r\n", HeaderWords[VEFCurRow], HeaderWords[VEFCurCol] );
			Printf( L"   Screen attribute   : %04x\r\n", HeaderWords[VEFColor] );
			Printf( L"   Window size        : %d rows x %d columns\r\n", HeaderWords[VEFWinBottom] - HeaderWords[VEFWinTop] + 1, HeaderWords[VEFWinRight] - HeaderWords[VEFWinLeft] + 1 );
			break;

		default :
			
			if ( !IgnoreNonSnapshotFiles )
				Printf( L"\r\n\"%s\" is not a Snapshot file.\r\n", Filename );
	}

	return 0;
}



int WildcardComparison2( LPTSTR Wild, LPTSTR Filename, int Ext, int Brackets )
{
	wchar_t temp[SHORT_BUF_LEN]				= L"";

	wchar_t *ch			= Wild;
	unsigned int j		= 0;
	int rv				= -99;


	while ( *ch ) {
		if ( *ch == '|' ) {
			if ( j ) {
				rv = WildcardComparison( temp, Filename, Ext, Brackets );
				if ( !rv )
					return rv;
			}

			j = 0;
			temp[j] = '\0';
		}
		else if ( j < ( SHORT_BUF_LEN - 1 ) ) {
			temp[j++] = *ch;
			temp[j] = '\0';
		}

		ch++;
	}

	if ( j )
		rv = WildcardComparison( temp, Filename, Ext, Brackets );

	return rv;	
}



int HandleMatchingFiles( LPTSTR Filespec, unsigned int Recurse, unsigned int Flags, unsigned int *TotalCount )
{
	//	wprintf( L"   * HandleMatchingFiles( '%s', %u )\n", Filespec, Flags );

	if ( !Filespec )
		return -1;
	else if ( Filespec[0] == '\0' )
		return -1;


	wchar_t Dir[MAX_FILENAME_LEN]			= L"";
	wchar_t Wild[MAX_FILENAME_LEN]			= L"";
	wchar_t SearchFor[MAX_FILENAME_LEN]		= L"";
	wchar_t FoundFull[MAX_FILENAME_LEN]		= L"";

	HANDLE Search			= NULL;
	WIN32_FIND_DATA Found;
	unsigned int FoundHere	= 0;
	BOOL ShowFooter			= TRUE;

	int trv					= 0;
	int rv					= 0;



	if ( Flags & SNAP_FLAG_NO_FOOTERS )
		ShowFooter = FALSE;

	if ( QueryIsDirectory( Filespec ) ) {
		wcscpy_s( Dir, MAX_FILENAME_LEN, Filespec );
		wcscpy_s( Wild, MAX_FILENAME_LEN, L"*" );
	}
	else {
		_PathPart( Filespec, Dir );
		_NamePart( Filespec, Wild );
	}

	wcscpy_s( SearchFor, MAX_FILENAME_LEN, Dir );
	AppendFilename( SearchFor, L"*", MAX_FILENAME_LEN );

	Search = FindFirstFile( SearchFor, &Found );

	if ( Search && ( Search != INVALID_HANDLE_VALUE ) ) {

		do {
			if ( Found.dwFileAttributes & ( FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_DEVICE | FILE_ATTRIBUTE_OFFLINE ) )
				continue;

			if ( WildcardComparison2( Wild, Found.cFileName, TRUE, TRUE ) )
				continue;

			FoundHere++;

			wcscpy_s( FoundFull, MAX_FILENAME_LEN, Dir );
			AppendFilename( FoundFull, Found.cFileName, MAX_FILENAME_LEN );

			trv = DisplaySnapInfo( FoundFull, ( Flags & SNAP_FLAG_IGNORE_OTHERS ) ? TRUE : FALSE );

			if ( trv > rv )
				rv = trv;
		}
		while ( FindNextFile( Search, &Found ) );

		FindClose( Search );
	}


	if ( !FoundHere && ( ( Recurse & RECURSE_FLAG_NO_FNF_ERROR ) == 0 ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No %s files found in %s\r\n", Wild, Dir );
	}
	else {
		if ( ( Flags & SNAP_FLAG_NO_ZERO_FOOTERS ) && !FoundHere )
			ShowFooter = FALSE;
		
		if ( ShowFooter )
			Printf( L"\r\nFound %u file%s in %s.\r\n", FoundHere, ( FoundHere == 1 ) ? L"" : L"s", Dir );
	}


	if ( Recurse & RECURSE_FLAG_RECURSE ) {

		Search = NULL;
		wcscpy_s( SearchFor, MAX_FILENAME_LEN, Dir );
		AppendFilename( SearchFor, L"*", MAX_FILENAME_LEN );

		Search = FindFirstFile( SearchFor, &Found );

		if ( Search && ( Search != INVALID_HANDLE_VALUE ) ) {

			do {
				if ( ( Found.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == 0 )
					continue;

				if ( !wcscmp( Found.cFileName, L"." ) || !wcscmp( Found.cFileName, L".." ) )
					continue;

				if ( Recurse & RECURSE_FLAG_NO_HIDDEN_DIRS )
					if ( Found.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN )
						continue;

				if ( Recurse & RECURSE_FLAG_NO_SYSTEM_DIRS )
					if ( Found.dwFileAttributes & FILE_ATTRIBUTE_SYSTEM )
						continue;

				if ( Recurse & RECURSE_FLAG_NO_JUNCTIONS )
					if ( Found.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT )
						continue;

				wcscpy_s( FoundFull, MAX_FILENAME_LEN, Dir );
				AppendFilename( FoundFull, Found.cFileName, MAX_FILENAME_LEN );
				AppendFilename( FoundFull, Wild, MAX_FILENAME_LEN );

				trv = HandleMatchingFiles( FoundFull, Recurse | RECURSE_FLAG_NO_FNF_ERROR, Flags | SNAP_FLAG_NO_ZERO_FOOTERS, TotalCount );

				if ( trv > rv )
					rv = trv;
			}
			while ( FindNextFile( Search, &Found ) );

			FindClose( Search );
		}
	}

	if ( TotalCount )
		*TotalCount += FoundHere;

	return rv;
}



LPTSTR QueryTrueName2( LPTSTR Filename, LPTSTR Output )
{
	wchar_t temp[MAX_FILENAME_LEN]		= L"";

	wchar_t *last			= temp;


	wcscpy_s( temp, MAX_FILENAME_LEN, Filename );

	if ( iswalnum( temp[0] ) && ( temp[1] == ':' ) )
		last += 2;

	for ( wchar_t *ch = temp; *ch; ch++ )
		if ( ( *ch == '\\' ) || ( *ch == '/' ) )
			last = ch;

	for ( wchar_t *ch = last; *ch; ch++ )
		if ( *ch == ';' )
			*ch = '|';

	return QueryTrueName( temp, Output );
}



int SnapFileInfo( unsigned int Flags, unsigned int Recurse )
{
	wchar_t Filespec[MAX_FILENAME_LEN]			= L"";

	unsigned int NumFilespecs	= 0;
	unsigned int TotalFound		= 0;

	int rv						= 0;


	for ( int i = 0; i < numargs; i++ ) {
		if ( arg[i][0] && ( arg[i][0] != '/' ) ) {
			Filespec[0] = '\0';
			QueryTrueName2( arg[i], Filespec );

			if ( Filespec[0] ) {
				if ( HandleMatchingFiles( Filespec, Recurse, Flags, &TotalFound ) )
					rv = 2;

				NumFilespecs++;
			}
			else {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Bad filespec:  \"%s\"\r\n", arg[i] );
				
				if ( rv < 2 )
					rv = 1;
			}
		}
	}

	if ( !NumFilespecs ) {
		Filespec[0] = '\0';
		QueryTrueName( L"*", Filespec );

		if ( HandleMatchingFiles( Filespec, Recurse, Flags, &TotalFound ) )
			rv = 2;

		NumFilespecs++;
	}

	if ( ( ( Flags & SNAP_FLAG_NO_FOOTERS ) == 0 ) && ( ( Recurse & RECURSE_FLAG_RECURSE ) || ( NumFilespecs > 1 ) ) )
		Printf( L"\r\nFound %u file%s total.\r\n", TotalFound, ( TotalFound == 1 ) ? L"" : L"s" );

	return rv;
}



int SetLastFilename( LPTSTR Filename )
{
	size_t len			= 0;


	if ( LastFilename )
		free( LastFilename );

	if ( !Filename ) {
		LastFilename = (LPTSTR) malloc( ( wcslen( ErrorString ) + 1 ) * sizeof( wchar_t ) );
		if ( !LastFilename )
			return -1;

		wcscpy_s( LastFilename, wcslen( ErrorString ) + 1, ErrorString );
	}
	else {
		len = wcslen( Filename ) + 3;

		LastFilename = (LPTSTR) malloc( len * sizeof( wchar_t ) );
		if ( !LastFilename )
			return -1;

		wcscpy_s( LastFilename, len, L"\"" );
		wcscat_s( LastFilename, len, Filename );
		wcscat_s( LastFilename, len, L"\"" );
	}

	return 0;
}



int CreateFilename( LPTSTR Filename, SYSTEMTIME SystemTime )
{
	if ( !Filename )
		return -1;


	wchar_t Base[MAX_FILENAME_LEN]		= L"";
	wchar_t temp[MAX_ARG_LEN]			= L"";
	wchar_t *NamePart					= NULL;


	if ( Filename[0] == '\0' )
		wcscpy_s( Filename, MAX_ARG_LEN, L".\\" );


	wsprintf( Base, L"Snapshot %04d-%02d-%02d %02d.%02d.%02d", SystemTime.wYear, SystemTime.wMonth, SystemTime.wDay, SystemTime.wHour, SystemTime.wMinute, SystemTime.wSecond );

	AppendFilename( Filename, Base, MAX_FILENAME_LEN );
	wcscat_s( Filename, MAX_FILENAME_LEN, L".snap" );

	if ( !QueryIsFileOrDirectory( Filename ) )
		return 0;


	NamePart = _NamePart( Filename, NULL );
	if ( !NamePart )
		NamePart = Filename;

	for ( unsigned int i = 0; i < 0x10000; i++ ) {
		*NamePart = '\0';
		wsprintf( temp, L"%s-%04X.snap", Base, i );
		wcscat_s( Filename, MAX_FILENAME_LEN, temp );
		if ( !QueryIsFileOrDirectory( Filename ) )
			return 0;
	}

	return -1;
}



int SaveSnapFile( LPTSTR Filename, unsigned int Flags )
{
	if ( !Filename )
		return 2;


	wchar_t ErrorText[SHORT_BUF_LEN]		= L"";

	int CreateMode					= CREATE_NEW;

	SYSTEMTIME Now;
	TIME_ZONE_INFORMATION TimeZoneInfo;
	CONSOLE_SCREEN_BUFFER_INFO ConsoleBufferInfo;
	UINT InputCodePage				= 0;
	UINT OutputCodePage				= 0;
	unsigned int TempBufferSize		= 0;

	HANDLE FileHandle				= NULL;
	HANDLE ConsoleHandle			= NULL;
	WORD *FileBuffer				= NULL;
	WORD *TempCharBuffer			= NULL;
	WORD *TempAttrBuffer			= NULL;
	_SnapHeader	Header;
	unsigned int BufferPtr			= 0;

	int rows						= 0;
	int cols						= 0;
	int endrow						= 0;
	int chunkrows					= 0;
	COORD BufCoords;
	DWORD WordsRead					= 0;
	DWORD BytesWritten				= 0;

	WORD data						= 0;
	WORD color						= 0;
	int count						= 0;

	int rv							= 2;



	//		Get the current time :

	GetLocalTime( &Now );
	GetTimeZoneInformation( &TimeZoneInfo );


	//		If no filename was specified, assume the current directory :

	if ( Filename[0] == '\0' ) {
		wcscpy_s( Filename, MAX_FILENAME_LEN, L"%_cwds" );
		ExpandVariables( Filename, 1 );
	}


	//		If the filename names a directory, generate a default filename :

	if ( QueryIsDirectory( Filename ) ) {
		if ( CreateFilename( Filename, Now ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error generating snapshot name in \"%s\"\r\n", Filename );
			return 2;
		}
	}


	//		Set file creation mode:

	if ( Flags & SNAP_FLAG_OVERWRITE )
		CreateMode = CREATE_ALWAYS;


	//		Open console handle :

	ConsoleHandle = CreateFile( L"CONOUT$", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL );

	if ( ConsoleHandle && ( ConsoleHandle != INVALID_HANDLE_VALUE ) ) {

		//		Get console info :

		if ( GetConsoleScreenBufferInfo( ConsoleHandle, &ConsoleBufferInfo ) ) {

			InputCodePage	= GetConsoleCP();
			OutputCodePage	= GetConsoleOutputCP();
			TempBufferSize	= ConsoleBufferInfo.dwSize.X * ConsoleBufferInfo.dwSize.Y;

			//		Allocate buffers :

			FileBuffer		= (WORD*) malloc( FILE_BUFFER_SIZE * sizeof( WORD ) );
			TempCharBuffer	= (WORD*) malloc( TempBufferSize * sizeof( WORD ) );
			TempAttrBuffer	= (WORD*) malloc( TempBufferSize * sizeof( WORD ) );

			if ( FileBuffer && TempCharBuffer && TempAttrBuffer ) {

				//		Open output file handle :

				FileHandle = CreateFile( Filename, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, CreateMode, FILE_ATTRIBUTE_NORMAL, NULL );

				if ( FileHandle && ( FileHandle != INVALID_HANDLE_VALUE ) ) {

					//		Create the file header :

					memset( &Header, 0x00, sizeof( Header ) );
					memcpy( Header.Signature, Signature, 16 );
					Header.Version		= 1;

					if ( ( Flags & SNAP_FLAG_NO_TIME_STAMPS ) == 0 ) {
						Header.Year			= Now.wYear;
						Header.Date			= ( ( Now.wMonth & 0x00ff ) << 8 ) | ( Now.wDay & 0x00ff );
						Header.Time			= ( ( Now.wHour & 0x00ff ) << 8 ) | ( Now.wMinute & 0x00ff );
						Header.Seconds		= ( ( Now.wDayOfWeek & 0x00ff ) << 8 ) | ( Now.wSecond & 0x00ff );
						Header.BiasLow		= TimeZoneInfo.Bias & 0x0000ffff;
						Header.BiasHigh		= TimeZoneInfo.Bias >> 16;
					}

					Header.NumColumns		= ConsoleBufferInfo.dwSize.X;
					Header.NumRows			= ConsoleBufferInfo.dwSize.Y;
					Header.CursorColumn		= ConsoleBufferInfo.dwCursorPosition.X;
					Header.CursorRow		= ConsoleBufferInfo.dwCursorPosition.Y;
					Header.Color			= ConsoleBufferInfo.wAttributes & 0x00ff;
					Header.CodePageInput	= InputCodePage;
					Header.CodePageOutput	= OutputCodePage;
					Header.WindowHeight		= ConsoleBufferInfo.srWindow.Bottom - ConsoleBufferInfo.srWindow.Top + 1;

					BytesWritten = 16;
					GetComputerName( Header.ComputerName, &BytesWritten );
					Header.ComputerName[15] = '\0';

					//		Copy the console buffer to the temp buffers :

					rows		= ConsoleBufferInfo.dwSize.Y;
					cols		= ConsoleBufferInfo.dwSize.X;
					chunkrows	= 8192 / cols;

					for ( int row = 0; row < rows; row += chunkrows ) {
						endrow = row + chunkrows - 1;
						if ( endrow >= rows )
							endrow = rows - 1;

						BufCoords.X		= 0;
						BufCoords.Y		= row;

						if ( !ReadConsoleOutputCharacter( ConsoleHandle, (LPTSTR) ( TempCharBuffer + ( row * cols ) ), ( endrow - row + 1 ) * cols, BufCoords, &WordsRead ) )
							goto CopyError;

						if ( !ReadConsoleOutputAttribute( ConsoleHandle, TempAttrBuffer + ( row * cols ), ( endrow - row + 1 ) * cols, BufCoords, &WordsRead ) )
							goto CopyError;
					}

					//		Checksum the raw data :

					Header.CRC32			= ChecksumTwoBuffers( TempCharBuffer, TempAttrBuffer, TempBufferSize );

					//		Write the header to the file :

					if ( !WriteFile( FileHandle, &Header, sizeof( _SnapHeader ), &BytesWritten, NULL ) )
						goto WriteError;

					//		Empty the output buffer :

					BufferPtr = 0;

					//		Loop through the buffers :

					color = ConsoleBufferInfo.wAttributes & 0x00ff;
					rv = 0;

					for ( unsigned int i = 0; i < TempBufferSize; i++ ) {
						data = TempCharBuffer[i];

						if ( TempAttrBuffer[i] != color ) {
							color = TempAttrBuffer[i];

							if ( rv = WriteWord( ColorCode, FileBuffer, &BufferPtr, FILE_BUFFER_SIZE, FileHandle ) )
								break;

							if ( rv = WriteWord( color, FileBuffer, &BufferPtr, FILE_BUFFER_SIZE, FileHandle ) )
								break;
						}

						count = 0;

						if ( ( i + 1 ) < TempBufferSize )
							if ( ( TempCharBuffer[i + 1] == data ) && ( TempAttrBuffer[i + 1] == color ) )
								count = RepeatCheck( TempCharBuffer, TempAttrBuffer, TempBufferSize, i );

						if ( ( count == -1 ) && ( data == SPACE ) ) {
							rv = WriteWord( EndOfDataCode, FileBuffer, &BufferPtr, FILE_BUFFER_SIZE, FileHandle );
							break;
						}
						else if ( ( count > 2 ) && ( data == SPACE ) ) {
							if ( rv = WriteWord( RepeatSpaceCode, FileBuffer, &BufferPtr, FILE_BUFFER_SIZE, FileHandle ) )
								break;

							if ( rv = WriteWord( count & 0xffff, FileBuffer, &BufferPtr, FILE_BUFFER_SIZE, FileHandle ) )
								break;

							i = i + count - 1;
						}
						else if ( count > 3 ) {
							if ( rv = WriteWord( RepeatCode, FileBuffer, &BufferPtr, FILE_BUFFER_SIZE, FileHandle ) )
								break;

							if ( rv = WriteWord( count & 0xffff, FileBuffer, &BufferPtr, FILE_BUFFER_SIZE, FileHandle ) )
								break;

							if ( rv = WriteWord( data, FileBuffer, &BufferPtr, FILE_BUFFER_SIZE, FileHandle ) )
								break;

							i = i + count - 1;
						}
						else if ( data >= 0xfff0 ) {
							if ( rv = WriteWord( EscapeCode, FileBuffer, &BufferPtr, FILE_BUFFER_SIZE, FileHandle ) )
								break;

							if ( rv = WriteWord( data, FileBuffer, &BufferPtr, FILE_BUFFER_SIZE, FileHandle ) )
								break;
						}
						else {
							if ( rv = WriteWord( data, FileBuffer, &BufferPtr, FILE_BUFFER_SIZE, FileHandle ) )
								break;
						}
					}

					//		Clean up and close the output file :

					if ( BufferPtr && !rv )
						rv = FlushBuffer( FileBuffer, &BufferPtr, FILE_BUFFER_SIZE, FileHandle );

					CloseHandle( FileHandle );

					if ( rv )
						goto WriteError;
				}
				else {
					GetLastErrorText( ErrorText );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Error opening \"%s\":\r\n", Filename );
					Qprintf( ERRHANDLE, L"    %s\r\n", ErrorText );
					rv = 2;
				}
			}
			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Can\u2019t allocate temporary buffers\r\n" );
				rv = 2;
			}

			free( FileBuffer );
			free( TempCharBuffer );
			free( TempAttrBuffer );
		}
		else {
			GetLastErrorText( ErrorText );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error getting console buffer info:\r\n" );
			Qprintf( ERRHANDLE, L"    %s\r\n", ErrorText );
			rv = 2;
		}

		CloseHandle( ConsoleHandle );
	}
	else {
		GetLastErrorText( ErrorText );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error opening console:\r\n" );
		Qprintf( ERRHANDLE, L"    %s\r\n", ErrorText );
		rv = 2;
	}


	if ( rv )
		SetLastFilename( NULL );
	else {
		SetLastFilename( Filename );

		if ( ( Flags & SNAP_FLAG_QUIET ) == 0 )
			PlaySound( L"CAMERA4", MyDLLHandle, SND_RESOURCE | SND_ASYNC );
	}

	return rv;


CopyError:


	GetLastErrorText( ErrorText );
	DisplayErrorHeader();
	Qprintf( ERRHANDLE, L"Error reading data from console buffer:\r\n" );
	Qprintf( ERRHANDLE, L"    %s\r\n", ErrorText );
	goto ErrorExit;


WriteError:


	GetLastErrorText( ErrorText );
	DisplayErrorHeader();
	Qprintf( ERRHANDLE, L"Error writing to file \"%s\":\r\n", Filename );
	Qprintf( ERRHANDLE, L"    %s\r\n", ErrorText );
	SetLastFilename( NULL );
	//	fall through....


ErrorExit:


	CloseHandle( FileHandle );
	CloseHandle( ConsoleHandle );
	free( FileBuffer );
	free( TempCharBuffer );
	free( TempAttrBuffer );

	return 2;
}



int RestoreSnapFile( LPTSTR Filename, unsigned int Flags )
{
	if ( Filename )
		if ( Filename[0] == '\0' )
			Filename = NULL;

	if ( !Filename ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename\r\n" );
		return 1;
	}


	wchar_t ErrorText[SHORT_BUF_LEN]			= L"";

	HANDLE ConsoleHandle			= NULL;
	HANDLE FileHandle				= NULL;
	WORD *FileBuffer				= NULL;
	WORD *TempCharBuffer			= NULL;
	WORD *TempAttrBuffer			= NULL;
	unsigned int TempBufferSize		= 0;
	DWORD BytesRead					= 0;

	unsigned int BufferPtr			= 0;
	unsigned int BufferValid		= 0;

	int FileType					= 0;
	_SnapHeader Header;

	WORD OldWinRows					= 0;
	WORD color						= 0;
	int data						= 0;
	int data2						= 0;
	unsigned int i					= 0;
	unsigned int count				= 0;
	BOOL vtoggle					= FALSE;

	HWND ConsoleWindowHandle		= NULL;
	CONSOLE_SCREEN_BUFFER_INFO CurrentBufferInfo;
	int CurrentCols					= 0;
	int CurrentRows					= 0;
	int NewCols						= 0;
	int NewRows						= 0;
	COORD ScreenBufferSize;
	SMALL_RECT ConsoleWinPos;

	int rows						= 0;
	int cols						= 0;
	int endrow						= 0;
	int chunkrows					= 0;
	COORD BufCoords;
	DWORD WordsWritten				= 0;
	COORD CursorPosition;

	int rv							= 0;



	//		Open console handle :

	ConsoleHandle = CreateFile( L"CONOUT$", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL );

	if ( ConsoleHandle && ( ConsoleHandle != INVALID_HANDLE_VALUE ) ) {

		//		Open file handle :

		FileHandle = CreateFile( Filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL );

		if ( FileHandle && ( FileHandle != INVALID_HANDLE_VALUE ) ) {

			//		Allocate file buffer :

			FileBuffer = (WORD*) malloc( FILE_BUFFER_SIZE * sizeof( WORD ) );

			if ( FileBuffer ) {

				//		Read header from file :

				memset( &Header, 0x00, sizeof( _SnapHeader ) );
				rv = ReadFile( FileHandle, FileBuffer, sizeof( _SnapHeader ), &BytesRead, NULL );

				if ( !rv || ( BytesRead != sizeof( _SnapHeader ) ) )
					goto ReadError;

				rv = 0;

				//		Check the file header :

				FileType = CheckHeader( (_SnapHeader*) FileBuffer );

				switch ( FileType ) {

				case 1 :

					memcpy( &Header, FileBuffer, sizeof( _SnapHeader ) );

					if ( Header.Version != 1 ) {
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Unsupported Snapshot file version:  %u\r\n", Header.Version );
						goto ErrorExit;
					}

					color		= Header.Color & 0x00ff;
					OldWinRows	= Header.WindowHeight;
					break;

				case 2 :

					memset( &Header, 0x00, sizeof( _SnapHeader ) );

					Header.NumColumns		= FileBuffer[VEFConCols];
					Header.NumRows			= FileBuffer[VEFConRows];
					Header.CursorColumn		= FileBuffer[VEFCurCol];
					Header.CursorRow		= FileBuffer[VEFCurRow];
					Header.Color			= FileBuffer[VEFColor];
					color					= Header.Color;
					Header.WindowHeight		= FileBuffer[VEFWinBottom] - FileBuffer[VEFWinTop] + 1;
					OldWinRows				= Header.WindowHeight;

					if ( FileBuffer[VEFHeaderEnd] == 0 )
						SetFilePointer( FileHandle, 0x0340, NULL, FILE_BEGIN );
					else
						SetFilePointer( FileHandle, 0x0016, NULL, FILE_BEGIN );

					break;

				default :

					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"File \"%s\" is not a Snapshot file.\r\n", Filename );
					goto ErrorExit;
				}

				//		Allocate temp buffers :

				TempBufferSize = Header.NumColumns * Header.NumRows;

				TempCharBuffer = (WORD*) malloc( TempBufferSize * sizeof( WORD ) );
				TempAttrBuffer = (WORD*) malloc( TempBufferSize * sizeof( WORD ) );

				if ( !TempCharBuffer || !TempAttrBuffer ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Can\u2019t allocate temporary buffers\r\n" );
					goto ErrorExit;
				}

				//		Empty the input buffer :

				BufferPtr	= 0;
				BufferValid	= 0;

				//		Read data from the file :

				i = 0;

				do {
					data = ReadWord( FileBuffer, &BufferPtr, &BufferValid, FILE_BUFFER_SIZE, FileHandle );

					if ( FileType == 2 ) {

						if ( data == -2 ) {
							if ( vtoggle )
								TempAttrBuffer[i++]	= Header.Color;

							while ( i < TempBufferSize ) {
								TempCharBuffer[i]	= SPACE;
								TempAttrBuffer[i++]	= Header.Color;
							}

							data = data2 = 0;
						}
						else if ( data < 0 )
							break;
						else {
							if ( vtoggle )
								TempAttrBuffer[i++] = data;
							else
								TempCharBuffer[i] = data;

							vtoggle = !vtoggle;
							continue;
						}
					}

					if ( data < 0 )
						break;

					if ( data == ColorCode ) {
						if ( ( data2 = ReadWord( FileBuffer, &BufferPtr, &BufferValid, FILE_BUFFER_SIZE, FileHandle ) ) < 0 )
							break;
						else
							color = data2;
					}
					else if ( ( data == RepeatCode ) || ( data == RepeatSpaceCode ) ) {
						if ( ( data2 = ReadWord( FileBuffer, &BufferPtr, &BufferValid, FILE_BUFFER_SIZE, FileHandle ) ) < 0 )
							break;
						else
							count = data2 ? data2 : 65536;

						if ( data == RepeatSpaceCode )
							data2 = SPACE;
						else {
							if ( ( data2 = ReadWord( FileBuffer, &BufferPtr, &BufferValid, FILE_BUFFER_SIZE, FileHandle ) ) < 0 )
								break;
						}

						for ( unsigned int j = 0; j < count; j++ ) {
							if ( i < TempBufferSize ) {
								TempCharBuffer[i]	= data2;
								TempAttrBuffer[i++]	= color;
							}
						}
					}
					else if ( data == EscapeCode ) {
						if ( ( data2 = ReadWord( FileBuffer, &BufferPtr, &BufferValid, FILE_BUFFER_SIZE, FileHandle ) ) < 0 )
							break;
						
						if ( i < TempBufferSize ) {
							TempCharBuffer[i]	= data2;
							TempAttrBuffer[i++]	= color;
						}
					}
					else if ( data == EndOfDataCode ) {
						while ( i < TempBufferSize ) {
							TempCharBuffer[i]	= SPACE;
							TempAttrBuffer[i++]	= Header.Color;
						}
					}
					else {
						if ( i < TempBufferSize ) {
							TempCharBuffer[i]	= data;
							TempAttrBuffer[i++]	= color;
						}
					}

				}
				while ( i < TempBufferSize );

				//		Done with the file :

				CloseHandle( FileHandle );
				free( FileBuffer );
				FileHandle	= NULL;
				FileBuffer	= NULL;

				//		Check for errors in file operations :

				if ( ( data < 0 ) || ( data2 < 0 ) )
					goto ReadError;

				//		Verify the checksum :

				if ( ( FileType == 1 ) && ( ( Flags & SNAP_FLAG_SKIP_CHECKSUM ) == 0 ) ) {
					if ( Header.CRC32 != ChecksumTwoBuffers( TempCharBuffer, TempAttrBuffer, TempBufferSize ) ) {
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"CRC32 mismatch\r\n" );
						goto ErrorExit;
					}
				}

				//		Get current console dimensions :

				memset( &CurrentBufferInfo, 0x00, sizeof( CONSOLE_SCREEN_BUFFER_INFO ) );
				ConsoleWindowHandle = GetConsoleWindow();
				if ( ConsoleWindowHandle )
					GetConsoleScreenBufferInfo( ConsoleHandle, &CurrentBufferInfo );

				//	Current size of the console window, in rows and columns:
				CurrentCols		= CurrentBufferInfo.srWindow.Right - CurrentBufferInfo.srWindow.Left + 1;
				CurrentRows		= CurrentBufferInfo.srWindow.Bottom - CurrentBufferInfo.srWindow.Top + 1;

				if ( ( CurrentCols < 2 ) || ( CurrentRows < 2 ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Can\u2019t get current screen buffer info\r\n" );
					goto ErrorExit;
				}

				//		Resize console window as needed :

				if ( ( CurrentBufferInfo.dwSize.X != Header.NumColumns ) || ( CurrentBufferInfo.dwSize.Y != Header.NumRows ) ) {

					if ( QueryIsTCMD() ) {
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Can\u2019t resize a Take Command tab window.\r\n" );
						Qprintf( ERRHANDLE, L"Detach the tab and try again.\r\n" );
						goto ErrorExit;
					}

					//		Resize the console window (preliminary) :

					NewCols		= ( Header.NumColumns < CurrentCols ) ? Header.NumColumns : CurrentCols;
					NewRows		= ( Header.WindowHeight < CurrentRows ) ? Header.WindowHeight : CurrentRows;

					if ( ( NewCols != CurrentCols ) || ( NewRows != CurrentRows ) ) {
						ResizeConsoleWindow( ConsoleHandle, ConsoleWindowHandle, NewCols, NewRows );
						CurrentCols	= NewCols;
						CurrentRows	= NewRows;
					}

					//		Set console screen buffer size :

					ScreenBufferSize.X	= Header.NumColumns;
					ScreenBufferSize.Y	= Header.NumRows;

					if ( !SetConsoleScreenBufferSize( ConsoleHandle, ScreenBufferSize ) ) {
						GetLastErrorText( ErrorText );
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Error resizing console screen buffer:\r\n" );
						Qprintf( ERRHANDLE, L"    %s\r\n", ErrorText );
						goto ErrorExit;
					}

					//		Resize the console window (final) :

					NewCols		= ( Header.NumColumns > CurrentCols ) ? Header.NumColumns : CurrentCols;
					NewRows		= ( Header.WindowHeight > CurrentRows ) ? Header.WindowHeight : CurrentRows;

					if ( ( NewCols != CurrentCols ) || ( NewRows != CurrentRows ) ) {
						ResizeConsoleWindow( ConsoleHandle, ConsoleWindowHandle, NewCols, NewRows );
						CurrentCols	= NewCols;
						CurrentRows	= NewRows;
					}
				}

				//		Copy temp buffers to console buffer :

				rows		= Header.NumRows;
				cols		= Header.NumColumns;
				chunkrows	= 8192 / cols;

				for ( int row = 0; row < rows; row += chunkrows ) {
					endrow = row + chunkrows - 1;
					if ( endrow > rows )
						endrow = rows - 1;

					BufCoords.X		= 0;
					BufCoords.Y		= row;

					if ( !WriteConsoleOutputCharacter( ConsoleHandle, (LPTSTR) ( TempCharBuffer + ( row * cols ) ), ( endrow - row + 1 ) * cols, BufCoords, &WordsWritten ) )
						goto CopyError;

					if ( !WriteConsoleOutputAttribute( ConsoleHandle, TempAttrBuffer + ( row * cols ), ( endrow - row + 1 ) * cols, BufCoords, &WordsWritten ) )
						goto CopyError;
				}

				//		Set cursor position :

				CursorPosition.X	= Header.CursorColumn;
				CursorPosition.Y	= Header.CursorRow;
				SetConsoleCursorPosition( ConsoleHandle, CursorPosition );

				//		Set console window position :

				ConsoleWinPos.Top		= Header.CursorRow - CurrentRows + 1;
				ConsoleWinPos.Bottom	= Header.CursorRow;
				ConsoleWinPos.Left		= 0;
				ConsoleWinPos.Right		= Header.NumColumns - 1;

				if ( ConsoleWinPos.Top < 0 ) {
					ConsoleWinPos.Top		= 0;
					ConsoleWinPos.Bottom	= CurrentRows - 1;
				}

				if ( ConsoleWinPos.Bottom >= Header.NumRows )
					ConsoleWinPos.Bottom = Header.NumRows;

				SetConsoleWindowInfo( ConsoleHandle, TRUE, &ConsoleWinPos );

				//		Restore miscellaneous data :

				SetConsoleTextAttribute( ConsoleHandle, Header.Color & 0x00ff );

				if ( FileType == 1 ) {
					SetConsoleOutputCP( Header.CodePageOutput );
					SetConsoleCP( Header.CodePageInput );
				}

				//		Done!

				free( TempCharBuffer );
				free( TempAttrBuffer );
			}
			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Can\u2019t allocate file buffer\r\n" );
				rv = 2;
			}

			CloseHandle( FileHandle );
		}
		else {
			GetLastErrorText( ErrorText );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error opening file \"%s\":\r\n", Filename );
			Qprintf( ERRHANDLE, L"    %s\r\n", ErrorText );
			rv = 2;
		}

		CloseHandle( ConsoleHandle );
	}
	else {
		GetLastErrorText( ErrorText );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error opening console:\r\n" );
		Qprintf( ERRHANDLE, L"    %s\r\n", ErrorText );
		rv = 2;
	}

	if ( !rv && ( ( Flags & SNAP_FLAG_QUIET ) == 0 ) )
		PlaySound( L"SystemNotification", NULL, SND_ALIAS | SND_ASYNC );

	return rv;


CopyError :

	GetLastErrorText( ErrorText );
	DisplayErrorHeader();
	Qprintf( ERRHANDLE, L"Error writing data to console buffer:\r\n" );
	Qprintf( ERRHANDLE, L"    %s\r\n", ErrorText );
	goto ErrorExit;


ReadError :


	GetLastErrorText( ErrorText );
	DisplayErrorHeader();
	Qprintf( ERRHANDLE, L"Error reading from file \"%s\":\r\n", Filename );
	Qprintf( ERRHANDLE, L"    %s\r\n", ErrorText );
	//	fall through...

ErrorExit :

	CloseHandle( FileHandle );
	CloseHandle( ConsoleHandle );
	free( FileBuffer );
	free( TempCharBuffer );
	free( TempAttrBuffer );
	return 2;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



DLLExports int WINAPI snap( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t Filename[MAX_FILENAME_LEN]		= L"";

	unsigned int Mode					= SNAP_MODE_SAVE;
	unsigned int Flags					= 0;
	unsigned int Recurse				= 0;
	LPTSTR SecondFilename				= NULL;

	int rv				= 0;


	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {

		if ( !wcscmp( arg[i], L"/?" ) )
			return ShowCmdHelp( SnapHelpText, FALSE );
		else if ( !wcscmp( arg[i], L"/??" ) )
			return ShowCmdHelp( SnapHelpText, TRUE );
		else if ( !_wcsicmp( arg[i], L"/E" ) )
			Recurse |= RECURSE_FLAG_NO_FNF_ERROR;
		else if ( !_wcsicmp( arg[i], L"/I" ) ) {
			if ( Mode == SNAP_MODE_INFO )
				Flags |= SNAP_FLAG_IGNORE_OTHERS;

			Mode = SNAP_MODE_INFO;
		}
		else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
			wchar_t *ch = arg[i] + 2;
			if ( IsColonOrEquals( *ch ) )
				ch++;

			while ( *ch ) {

				switch ( towupper( *ch ) ) {

				case 'D' :

					Recurse |= RECURSE_FLAG_NO_HIDDEN_DIRS;
					break;

				case 'E' :

					Recurse |= RECURSE_FLAG_NO_FNF_ERROR;
					break;

				case 'J' :

					Recurse |= RECURSE_FLAG_NO_JUNCTIONS;
					break;

				case 'O' :

					Flags |= SNAP_FLAG_IGNORE_OTHERS;
					break;

				case 'S' :

					Flags |= SNAP_FLAG_NO_FOOTERS;
					break;

				case 'T' :

					Flags |= SNAP_FLAG_NO_TIME_STAMPS;
					break;

				case 'X' :

					Flags |= SNAP_FLAG_SKIP_CHECKSUM;
					break;

				case 'Z' :

					Recurse |= RECURSE_FLAG_NO_SYSTEM_DIRS;
					break;

				default :

					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Unknown suboption:  /N:%c\r\n", towupper( *ch ) );
					return 1;
				}

				ch++;
			}
		}
		else if ( !_wcsicmp( arg[i], L"/O" ) )
			Flags |= SNAP_FLAG_OVERWRITE;
		else if ( !_wcsicmp( arg[i], L"/Q" ) )
			Flags |= SNAP_FLAG_QUIET;
		else if ( !_wcsicmp( arg[i], L"/R" ) )
			Mode = SNAP_MODE_RESTORE;
		else if ( !_wcsicmp( arg[i], L"/S" ) )
			Recurse |= RECURSE_FLAG_RECURSE;
		else if ( arg[i][0] == '/' ) {
			_wcsupr_s( arg[i], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Unknown option:  \"%s\"\r\n", arg[i] );
			return 1;
		}
		else if ( Filename[0] == '\0' ) {
			QueryTrueName( arg[i], Filename );

			if ( Filename[0] == '\0' ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad filename:  \"%s\"\r\n", arg[i] );
				return 1;
			}
		}
		else
			SecondFilename = arg[i];
	}


	if ( SecondFilename && ( Mode != SNAP_MODE_INFO ) ) {
		_wcsupr_s( SecondFilename, MAX_ARG_LEN );
		Qprintf( ERRHANDLE, L"Syntax error:  \"%s\"\r\n", SecondFilename );
		return 1;
	}


	if ( SilencedByRegistry() )
		Flags |= SNAP_FLAG_QUIET;


	switch ( Mode ) {

	case SNAP_MODE_INFO :

		rv = SnapFileInfo( Flags, Recurse );
		break;

	case SNAP_MODE_RESTORE :

		rv = RestoreSnapFile( Filename, Flags );
		break;

	default :

		rv = SaveSnapFile( Filename, Flags );
	}

	return rv;
}



DLLExports int WINAPI _lastsnap( LPTSTR outString )
{
	if ( !outString )
		return 1;


	if ( LastFilename )
		wcscpy_s( outString, MAX_FILENAME_LEN, LastFilename );
	else
		outString[0] = '\0';

	return 0;
}



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	MyDLLHandle = (HINSTANCE) hModule;

	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/snapshot.html";
	piInfo.pszDescription	= (LPTSTR) L"Save and restore the console screen buffer";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	if ( wcscmp( shortbuf, L"00" ) == 0 )
		Printf( L"%s v%d.%d.%d loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	if ( LastFilename )
		free( LastFilename );

	return 0;
}



const int NumHelpTexts							= 2;
const wchar_t HelpFeature[NumHelpTexts][11]		= { L"snap", L"_lastsnap" };
const wchar_t *HelpText[NumHelpTexts]			= { SnapHelpText, _LastSnapHelpText };



LPTSTR UsageText( const wchar_t *inString )
{
	if ( !inString )
		return NULL;


	wchar_t *ch = wcschr( (wchar_t*) inString, '\n' );


	if ( !ch )
		return NULL;

	while ( *ch == '\n' )
		ch++;

	return ch;
}



DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( ( inString[0] == '@' ) || ( inString[0] == '_' ) )
		return 0;

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			wcscpy_s( outString, SHORT_BUF_LEN, UsageText( HelpText[i] ) );
			return 1;
		}
	}

	return 0;
}



DLLExports int WINAPI Help( LPTSTR inString )
{
	wchar_t WindowTitle[SHORT_BUF_LEN]		= L"";


	wsprintf( WindowTitle, L"%s Plugin Help", PLUGIN_NAME );

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			MessageBox( GetForegroundWindow(), HelpText[i], WindowTitle, MB_OK );
			return 1;
		}
	}

	return 0;
}
